/*
 * ********************************************************************************************************************
 *
 *               iFHS7.
 *              ;BBMBMBMc                  rZMBMBR              BMB
 *              MBEr:;PBM,               7MBMMEOBB:             BBB                       RBW
 *     XK:      BO     SB.     :SZ       MBM.       c;;     ir  BBM :FFr       :SSF:    ;xBMB:r   iuGXv.    i:. iF2;
 *     DBBM0r.  :D     S7   ;XMBMB       GMBMu.     MBM:   BMB  MBMBBBMBMS   WMBMBMBBK  MBMBMBM  BMBRBMBW  .MBMBMBMBB
 *      :JMRMMD  ..    ,  1MMRM1;         ;MBMBBR:   MBM  ;MB:  BMB:   MBM. RMBr   sBMH   BM0         UMB,  BMB.  KMBv
 *     ;.   XOW  B1; :uM: 1RE,   i           .2BMBs  rMB. MBO   MBO    JMB; MBB     MBM   BBS    7MBMBOBM:  MBW   :BMc
 *     OBRJ.SEE  MRDOWOR, 3DE:7OBM       .     ;BMB   RMR7BM    BMB    MBB. BMB    ,BMR  .BBZ   MMB   rMB,  BMM   rMB7
 *     :FBRO0D0  RKXSXPR. JOKOOMPi       BMBSSWBMB;    BMBB:    MBMB0ZMBMS  .BMBOXRBMB    MBMDE RBM2;SMBM;  MBB   xBM2
 *         iZGE  O0SHSPO. uGZ7.          sBMBMBDL      :BMO     OZu:BMBK,     rRBMB0;     ,EBMB  xBMBr:ER.  RDU   :OO;
 *     ,BZ, 1D0  RPSFHXR. xWZ .SMr                  . .BBB
 *      :0BMRDG  RESSSKR. 2WOMBW;                   BMBMR
 *         i0BM: SWKHKGO  MBDv
 *           .UB  OOGDM. MK,                                          Copyright (c) 2015-2019.  斯伯坦机器人
 *              ,  XMW  ..
 *                  r                                                                     All rights reserved.
 *
 * ********************************************************************************************************************
 */

package com.sybotan.android.base.comparators

import java.io.File

/**
 * 文件排序器
 *
 * @author  庞利祥(sybotan@126.com)
 */
class SFileComparator {
    /**
     * 名称升序比较器
     *
     * @author  庞利祥(sybotan@126.com)
     */
    class NameAsc : Comparator<File> {
        /**
         * 文件名比较
         *
         * @param   file1   文件1
         * @param   file2   文件2
         * @return  如果文件1比文件2名称靠后返回1，否则返回-1
         */
        override fun compare(file1: File, file2: File): Int {
            return if (file1.name > file2.name) {
                1
            } else {
                -1
            }
        } // Function compare()
    } // LastModifiedAsc

    /**
     * 文件名比较器
     *
     * @author  庞利祥(sybotan@126.com)
     */
    class NameDesc : Comparator<File> {
        /**
         * 创建时间比较
         *
         * @param   file1   文件1
         * @param   file2   文件2
         * @return  如果文件1比文件2名称靠前返回1，否则返回-1
         */
        override fun compare(file1: File, file2: File): Int {
            return if (file1.name < file2.name) {
                1
            } else {
                -1
            }
        } // Function compare()
    } // NameDesc

    /**
     * 创建时间升序比较器
     *
     * @author  庞利祥(sybotan@126.com)
     */
    class LastModifiedAsc : Comparator<File> {
        /**
         * 创建时间比较
         *
         * @param   file1   文件1
         * @param   file2   文件2
         * @return  如果文件1比文件2修改时间靠后返回1，否则返回-1
         */
        override fun compare(file1: File, file2: File): Int {
            return if (file1.lastModified() > file2.lastModified()) {
                1
            } else {
                -1
            }
        } // Function compare()
    } // LastModifiedAsc

    /**
     * 创建时间降序比较器
     *
     * @author  庞利祥(sybotan@126.com)
     */
    class LastModifiedDesc : Comparator<File> {
        /**
         * 创建时间比较
         *
         * @param   file1   文件1
         * @param   file2   文件2
         * @return  如果文件1比文件2修改时间靠前返回1，否则返回-1
         */
        override fun compare(file1: File, file2: File): Int {
            return if (file1.lastModified() < file2.lastModified()) {
                1
            } else {
                -1
            }
        } // Function compare()
    } // LastModifiedDesc
} // Class SFileComparator